#include <array>
#include <algorithm> // 'std::copy_if'...
#include <iostream>
#include <utility> // 'std::integer_sequence'...

#include "Constexpr.h"

template<int... Integers> // C++14
constexpr std::array<int, sizeof...(Integers)> create_array(std::integer_sequence<int, Integers...>) noexcept {
	return std::array<int, sizeof...(Integers)> { Integers... }; // Aggregate initialization...
}
constexpr std::array<int, 20> My_array = ::create_array(std::make_integer_sequence<int, 20>{});
static_assert(My_array[19] == 19);

// This requires the explicit introduction of 'Result' at instantiation time:
template<typename T, typename Result> Result multiply(T t1, T t2) {
	return t1 * t2;
}
// Compiler fails because 't1' and 't2' are undefined (left-to-right parsing):
//template<typename T> decltype(t1 * t2) multiply(T t1, T t2) {
//	decltype(t1 * t2) t = t1 * t2; // This works here only...
//	return t;
//}
// Trailing return type thanks to 'auto':
template<typename T> /* [] */ auto multiply(T t1, T t2) -> decltype(t1* t2) {
	return t1 * t2;
}

/** Inspiration from: */
// https://stackoverflow.com/questions/19019252/create-n-element-constexpr-array-in-c11/34465458

constexpr int Not_prime = -1;

constexpr int As_prime(int n) {
	if (n < 2) return ::Not_prime;
	for (int i = 2; i < n; i++)
		if (n % i == 0)
			return ::Not_prime;
	return n;
}

// 'std::index_sequence' and 'std::make_index_sequence' are specializations of
// 'std::integer_sequence' and 'std::make_integer_sequence' with 'T' = 'std::size_t'
/** Internal '_create_array' template function called by 'create_array' */
template<typename Function, std::size_t... Integers>
// Trailing return type thanks to 'auto':
constexpr/*[]*/ auto _create_array(Function f, std::index_sequence<Integers...>) ->
// Returned type:
std::array<std::invoke_result_t<Function, std::size_t>, sizeof...(Integers)> {
	return { { f(Integers)... } }; // Aggregate initialization
}

template<int N, typename Function>
// Trailing return type based on 'auto':
constexpr/*[]*/ auto create_array(Function f) ->
// Returned type:
std::array<typename std::result_of<Function(std::size_t)>::type, N> {
	// 'std::make_index_sequence<7>{}' -> '<0,1,2,3,4,5,6,7>' become 'Integers'
	return _create_array<Function>(f, std::make_index_sequence<N>{});
}

constexpr std::array<int, Constexpr_example::N> Primes =
::create_array<Constexpr_example::N/*, decltype(&::As_prime)*/>(::As_prime);
static_assert(Primes[0] == ::Not_prime); // No compilation error
static_assert(Primes[1] == ::Not_prime); // No compilation error
static_assert(Primes[2] != ::Not_prime); // No compilation error
static_assert(Primes[3] != ::Not_prime); // No compilation error

int main() {
	// Non-prime numbers are substituted by '::Not_prime'. So display only primes such that 'p' < 'N':
	std::copy_if(std::begin(::Primes), std::end(::Primes), std::ostream_iterator<int>(std::cout, " - "),
		[](int i) -> bool {
			return i != ::Not_prime;
		});

	/** Code is put inside 'main' for illustration purpose only! */
	// 'N' is made available in local scope ('main') for *CAPTURE*;
	constexpr int N = Constexpr_example::N;
	// 'constexpr' lambda expression:
	constexpr auto my_lambda = [N /* Capture 'constexpr' only */]() {
		if (::Is_prime<Constexpr_example::N>()) // '::Is_prime<N>()' *DOES NOT* work...
			return ::As_prime(N); // '::As_prime(Constexpr_example::N)' *DOES* work...
		return ::Not_prime;
	};
	static_assert(my_lambda() == ::Not_prime); // No compilation error since 'Constexpr_example::N == 8'
}